"""oas_dataservice module for mapping a openapi-specification to dcat rdf.

This module contains methods for mapping an openAPI specification to rdf
dcat:DataService according to the
`dcat-ap-no v.2 standard <https://doc.difi.no/review/dcat-ap-no/#klasse-dataset>`__

Example:
    >>> import yaml
    >>> import requests
    >>> from datacatalogtordf import Catalog
    >>> from oastodcat import OASDataService
    >>>
    >>> # Create catalog object
    >>> catalog = Catalog()
    >>> catalog.identifier = "http://example.com/catalogs/1"
    >>> catalog.title = {"en": "A dataset catalog"}
    >>> catalog.publisher = "https://example.com/publishers/1"
    >>>
    >>> # Create a dataservice based on an openAPI-specification:
    >>> url = ("https://raw.githubusercontent.com/"
    >>>        "OAI/OpenAPI-Specification/master/examples/v3.0/petstore.yaml"
    >>>       )
    >>> oas = yaml.safe_load(requests.get(url).text)
    >>> identifier = "http://example.com/dataservices/{id}"
    >>> oas_spec = OASDataService(url, oas, identifier)
    >>> #
    >>> # Add dataservices to catalog:
    >>> for dataservice in oas_spec.dataservices:
    >>>    catalog.services.append(dataservice)
    >>>
    >>> # get dcat representation in turtle (default)
    >>> dcat = catalog.to_rdf()
    >>> print(dcat.decode())
    >>> bool(dcat)
    True
"""
import hashlib
from typing import List, Optional

from concepttordf import Contact
from datacatalogtordf import DataService, URI


class OASDataService:
    """A simple class representing an openAPI specification.

    When initialized, the specification is parsed and one or more instances
    of dcat:DataService is added to the list dataservices.

    Attributes:
        specification (dict): an openAPI spec as a dict
        dataservices (List[DataService]): a list of dataservices created
        endpointdescription (str): The url of the openAPI specification
        identifier (str): the identifier template, should contain {id}
    """

    __slots__ = (
        "_specification",
        "_dataservices",
        "_identifier",
        "_endpointdescription",
        "_media_types",
        "_dataservice",
        "_publisher",
        "_conforms_to",
    )

    # Types:
    _specification: dict
    _dataservices: List[DataService]
    _identifier: str
    _endpointdescription: URI
    _media_types: List[str]
    _publisher: str
    _conforms_to: List[str]

    def __init__(self, url: str, specification: dict, identifier: str) -> None:
        """Inits an object with default values and parses the specification.

        Args:
            url (str): the url of the openAPI specification
            specification (dict): an openAPI specification as a dict
            identifier (str): the identifier template, containing {id}

        Raises:
            NotSupportedOASError: We do not support this version of the specification
            NotValidOASError: The specification is not valid
            RequiredFieldMissingError: a required property is missing
        """
        super().__init__()
        if not (specification):
            raise NotValidOASError("Empty specification object")

        if not specification["openapi"].startswith("3.0."):
            raise NotSupportedOASError(
                f'Version {specification["openapi"]}" is not supported'
            )
        if len(identifier) == 0:
            raise RequiredFieldMissingError("Empty indentification attribute")

        self.identifier = identifier
        self.endpointdescription = url
        self.specification = specification
        self._dataservices: List[DataService] = []
        self._media_types = []

        # endpointURL
        if "servers" in specification:
            for server in specification["servers"]:
                if "url" in server:
                    self._create_dataservice(url=server["url"])
        else:
            self._create_dataservice()

    @property
    def identifier(self) -> str:
        """Get/set for identifier."""
        return self._identifier

    @identifier.setter
    def identifier(self, identifier: str) -> None:
        self._identifier = identifier

    @property
    def endpointdescription(self) -> str:
        """Get/set for endpointdescription."""
        return self._endpointdescription

    @endpointdescription.setter
    def endpointdescription(self, endpointdescription: str) -> None:
        self._endpointdescription = URI(endpointdescription)

    @property
    def specification(self) -> dict:
        """Get/set for specification."""
        return self._specification

    @specification.setter
    def specification(self, specification: dict) -> None:
        self._specification = specification

    @property
    def publisher(self) -> str:
        """Get/set for publisher."""
        return self._publisher

    @publisher.setter
    def publisher(self, publisher: str) -> None:
        self._publisher = URI(publisher)
        for dataservice in self._dataservices:
            dataservice.publisher = publisher

    @property
    def conforms_to(self) -> List[str]:
        """Get/set for conforms_to."""
        return self._conforms_to

    @conforms_to.setter
    def conforms_to(self, conforms_to: List[str]) -> None:
        self._conforms_to = conforms_to
        for dataservice in self._dataservices:
            dataservice.conformsTo = conforms_to

    @property
    def dataservices(self) -> List[DataService]:
        """Get for dataservices."""
        return self._dataservices

    # --
    def _create_dataservice(self, url: Optional[str] = None) -> None:
        """Creates a dataservice instance and appends it to list of dataservices."""
        self._dataservice = DataService()
        if url:
            self._dataservice.endpointURL = url
        self._dataservice.endpointDescription = self.endpointdescription

        try:
            self._dataservice.publisher = self.publisher
        except AttributeError:
            pass

        try:
            self._dataservice.conformsTo = self.conforms_to
        except AttributeError:
            self.conforms_to: List[str] = []

        self._parse_specification()

        # We may be given an identifier "template" ending with {id}.
        # We create the identifier and url based on title and complete the identifer:
        id = (
            self._dataservice.title["en"]
            if url is None
            else self._dataservice.title["en"] + url
        )
        self._dataservice.identifier = URI(self.identifier.format(id=create_id(id)))

        self.dataservices.append(self._dataservice)

    def _parse_specification(self) -> None:
        """Parses the specification and adds the attribute to the dataservice."""
        # title
        self._parse_title()
        # description
        self._parse_description()
        # contactpoint
        self._parse_contactpoint()
        # license
        self._parse_license()
        # mediaType
        self._parse_media_type()
        # externalDocs
        self._parse_external_docs()

    def _parse_title(self) -> None:
        """Parses the title object."""
        if "title" in self.specification["info"]:
            # Assuming English
            # title
            self._dataservice.title = {"en": self.specification["info"]["title"]}

    def _parse_description(self) -> None:
        """Parses the description object."""
        if "description" in self.specification["info"]:
            self._dataservice.description = {
                "en": self.specification["info"]["description"]
            }

    def _parse_contactpoint(self) -> None:
        """Parses the contact object."""
        if "contact" in self.specification["info"]:
            contact = Contact()
            if "name" in self.specification["info"]["contact"]:
                contact.name = {"en": self.specification["info"]["contact"]["name"]}
            if "email" in self.specification["info"]["contact"]:
                contact.email = self.specification["info"]["contact"]["email"]
            if "url" in self.specification["info"]["contact"]:
                contact.url = self.specification["info"]["contact"]["url"]
            self._dataservice.contactpoint = contact

    def _parse_license(self) -> None:
        """Parses the license object."""
        if "license" in self.specification["info"]:
            if "url" in self.specification["info"]["license"]:
                self._dataservice.license = self.specification["info"]["license"]["url"]

    def _parse_media_type(self) -> None:
        """Parses the media type objects."""
        self._seek_media_types(self.specification, ["content"])
        # Need to remove duplicates:
        self._dataservice.media_types = list(set(self._media_types))

    def _parse_external_docs(self) -> None:
        """Parses the externalDocs objects."""
        if "externalDocs" in self.specification:
            if "url" in self.specification["externalDocs"]:
                self._dataservice.landing_page.append(
                    self.specification["externalDocs"]["url"]
                )

    # --
    def _seek_media_types(self, d: dict, key_list: List[str]) -> None:
        """Helper method.

        Seeks for keys matching any of keys in key_list.
        Adds matching keys to self._media_types.

        Args:
            d (dict): the dict in which to searc
            key_list (List[str]): list of keys to search for
        """
        _url = "https://www.iana.org/assignments/media-types/"
        for k, v in d.items():
            if k in key_list:
                for key in v.keys():
                    self._media_types.append(_url + str(key))
            if isinstance(v, dict):
                self._seek_media_types(v, key_list)


def create_id(s: str) -> str:
    """Helper function to create unique ids based on input str s."""
    return hashlib.sha1(str.encode(s)).hexdigest()  # noqa: S303


class Error(Exception):
    """Base class for exceptins in this module."""

    pass


class NotValidOASError(Error):
    """The specification object is not valid.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message: str) -> None:
        """Inits an object with default values."""
        self.message = message


class NotSupportedOASError(Error):
    """We do not support this version of the specification.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message: str) -> None:
        """Inits an object with default values."""
        self.message = message


class RequiredFieldMissingError(Error):
    """A required filed is missing.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message: str) -> None:
        """Inits an object with default values."""
        self.message = message



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['oastodcat']

package_data = \
{'': ['*']}

install_requires = \
['datacatalogtordf>=1.0.0,<2.0.0',
 'pyyaml>=5.3.1,<6.0.0',
 'rdflib-jsonld>=0.5.0,<0.6.0',
 'rdflib>=5.0.0,<6.0.0',
 'requests>=2.24.0,<3.0.0']

extras_require = \
{':python_version < "3.8"': ['importlib_metadata>=1.5.0,<2.0.0']}

setup_kwargs = {
    'name': 'oastodcat',
    'version': '2.0.0a4',
    'description': 'A library for transforming an openAPI file to one or more dcat:DataService',
    'long_description': '![Tests](https://github.com/Informasjonsforvaltning/oastodcat/workflows/Tests/badge.svg)\n[![codecov](https://codecov.io/gh/Informasjonsforvaltning/oastodcat/branch/master/graph/badge.svg)](https://codecov.io/gh/Informasjonsforvaltning/oastodcat)\n[![PyPI](https://img.shields.io/pypi/v/oastodcat.svg)](https://pypi.org/project/oastodcat/)\n[![Read the Docs](https://readthedocs.org/projects/oastodcat/badge/)](https://oastodcat.readthedocs.io/)\n# oastodcat\n\nA small Python library to transform an openAPI file to a dcat:DataService\n\nAt this moment we support all 3.0.x versions of (The OpenAPI specification)[https://github.com/OAI/OpenAPI-Specification]\n\n## Usage\n### Install\n```\n% pip install oastodcat\n```\n### Getting started\nExample usage:\n```\nimport yaml\nimport requests\nfrom datacatalogtordf import Catalog\nfrom oastodcat import OASDataService\n\n# Create catalog object\ncatalog = Catalog()\ncatalog.identifier = "http://example.com/catalogs/1"\ncatalog.title = {"en": "A dataset catalog"}\ncatalog.publisher = "https://example.com/publishers/1"\n\n# Create a dataservice based on an openAPI-specification:\nurl = ("https://raw.githubusercontent.com/"\n      "OAI/OpenAPI-Specification/master/examples/v3.0/petstore.yaml"\n     )\noas = yaml.safe_load(requests.get(url).text)\nidentifier = "http://example.com/dataservices/{id}"\noas_spec = OASDataService(url, oas, identifier)\n#\n# Add dataservices to catalog:\nfor dataservice in oas_spec.dataservices:\n  catalog.services.append(dataservice)\n\nget dcat representation in turtle (default)\ndcat = catalog.to_rdf()\n# Get dcat representation in turtle (default)\ndcat = catalog.to_rdf()\nprint(dcat.decode())\n```\n\n## Mapping\nThe following table shows how an openAPI specification is mapped to a dcat:DataService:  \n(Only dcat:DataService properties are shown.)\n\n| dcat:DataService         | RDF property             | openAPI v 3.0.x      | Note |\n| ------------------------ | ------------------------ | -------------------- | ---- |\n| endpoint description     | dcat:endpointDescription | <url to description> |      |\n| endpoint URL             | dcat:endpointURL         | servers.url          | [1]  |\n| serves dataset           |                          | _n/a_                |      |\n| access rights            |                          |                      |      |\n| conforms to              | dct:conformsTo           |                      |      |\n| contact point            | dcat:contactPoint        | info.contact         |      |\n| creator                  |                          |                      |      |\n| description              | dct:description          | info.description     |      |\n| has policy               |                          |                      |      |\n| identifier               |                          | _n/a_                |      |\n| is referenced by         |                          |                      |      |\n| keyword/tag              |                          |                      |      |\n| landing page             | dcat:landingPage         | externalDocs         |      |\n| license                  | dct:license              | info.license.url     |      |\n| resource language        |                          |                      |      |\n| relation                 |                          |                      |      |\n| rights                   |                          |                      |      |\n| qualified relation       |                          |                      |      |\n| publisher                | dct:publisher            |                      |      |\n| release date             |                          |                      |      |\n| theme/category           |                          |                      |      |\n| title                    | dct:title                | info.title           |      |\n| type/genre               |                          |                      |      |\n| update/modification date |                          |                      |      |\n| qualified attribution    |                          |                      |      |\n| _media type_             | dcat:mediaType           | <it\'s complicated>   |      |\n\n[1] For each url in the servers object array, an instance of dcat:DataService will be created.\n\n## Development\n### Requirements\n- python3\n- [pyenv](https://github.com/pyenv/pyenv) (recommended)\n- [pipx](https://github.com/pipxproject/pipx) (recommended)\n- [poetry](https://python-poetry.org/)\n- [nox](https://nox.thea.codes/en/stable/)\n- [nox-poetry](https://pypi.org/project/nox-poetry/)\n```\n% pipx install poetry==1.0.5\n% pipx install nox==2019.11.9\n% pipx inject nox nox-poetry\n```\n\n### Install\n```\n% git clone https://github.com/Informasjonsforvaltning/oastodcat.git\n% cd oastodcat\n% pyenv install 3.8.6\n% pyenv install 3.7.9\n% pyenv install 3.9.0\n% pyenv local 3.9.0 3.8.3 3.7.7\n% poetry install\n```\n### Run all sessions\n```\n% nox\n```\n### Run all tests with coverage reporting\n```\n% nox -rs tests\n```\n### Debugging\nYou can enter into [Pdb](https://docs.python.org/3/library/pdb.html) by passing `--pdb` to pytest:\n```\nnox -rs tests -- --pdb\n```\nYou can set breakpoints directly in code by using the function `breakpoint()`.\n',
    'author': 'Stig B. Dørmænen',
    'author_email': 'stigbd@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Informasjonsforvaltning/oastodcat',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
